#!/usr/bin/perl
# Adam's method of calculating Frisch/Broe/Pierrehumbert similarity
#
#use bytes;
# if there's an argument, use that as the name of the feature file

if (@ARGV > 0) {
    $featurefile = $ARGV[0];
} else {
    print "Enter name of feature file: ";
    $featurefile = <STDIN>;
    chomp($featurefile);
}

open( FEATUREFILE, "<$featurefile")  or die "Can't open input file: $!\n";

print "Select type of natural class descriptions to use:\n\t1. Contrastive underspecification (e.g., [u] = [+back, +high])\n\t2. Fully specified (e.g., [u] = [+back, +high, -low, +round])\n(The similarity results will be the same either way)\n? ";
while ($continue == 0) {
    $specification = <STDIN>;
    chomp($specification);
    if ($specification eq "1") {
	$specification = "under";
	$continue = 1;		
    } elsif ($specification eq "2") {
        $specification = "over";        
	$continue = 1;		
    } else {
        print "Sorry, I couldn't understand your response.  Please enter 1 or 2.\n?  ";	
    }
}
$continue = undef;

print "\nInclude the maximal superclass? (That is, the class that includes all known segments)\n\t1. Automatically include (a la Frisch)\n\t2. Exclude if not a genuine class (a la Zuraw) \n(These options yield slightly different absolute similarity values, but the same relative similarities)\n? ";
while ($continue == 0) {
    $superclass = <STDIN>;
    chomp($superclass);
    if ($superclass eq "1") {
	$superclass = "include";
	$continue = 1;		
    } elsif ($superclass eq "2") {
        $superclass = "exclude";        
	$continue = 1;		
    } else {
        print "Sorry, I couldn't understand your response.  Please enter 1 or 2.\n?  ";	
    }
}
$continue = undef;


$filename = $featurefile;
# we'll take off a suffix to create a default filename for the output files
$filename =~ s/\.[^\.]*$//;
open ( OUTPUTFILE, ">$filename.log") or die "Can't open output file: $!\n";
$similarity_table_filename = "$filename.stb";
$natural_class_filename = "$filename.cls";
$sim_matrix_filename = "$filename.sim";
$trans_filename = "$filename.trans";

# check to make sure there isn't already a file with this name
while ($continue == 0) {
    if (-e $similarity_table_filename or -e $natural_class_filename) {
	# there's already a file with this name; ask if OK to overwrite
	print "\nFile $similarity_table_filename, $natural_class_filename, or $sim_matrix_filename already exists.\nWhat should I do?\n\t1. overwrite old $similarity_table_filename\n\t2. Save similarity table under a new name.\n? ";
	$response = <STDIN>;
	chomp($response);
	if ($response eq "1" or $response eq "1.") {
	    # OK to overwrite; break out and keep going
	    $continue = 1;
	} elsif ($response eq "2" or $response eq "2.") { 
	    print "Enter new name for similarity table file: ";
	    $similarity_table_filename = <STDIN>;
	    chomp($similarity_table_filename);
	    $natural_class_filename = $similarity_table_filename;
	    $sim_matrix_filename = $similarity_table_filename;

	    # if new filename already ends in .stb, it's OK; otherwise add it
	    if ($similarity_table_filename =~ /.stb$/) {
		$natural_class_filename =~ s/.stb$/.cls/;
		$sim_matrix_filename =~ s/.stb$/.sim/;
	    } else {
		$similarity_table_filename .= ".stb";
		$natural_class_filename .= ".cls";
		$sim_matrix_filelname .= ".sim";
	    }
	    # we won't set $continue to 1 yet, so we can check if the
	    # new filename also already exists.
	} else {
	    print "Sorry, couldn't understand response; please try again.\n";
	}
    } else {
	# file doesn't already exist.  proceed normally.
	$continue = 1;
    }
}
$continue = undef;

open ( SIMMATRIXFILE, ">$sim_matrix_filename") or die "Can't write similarity matrix file: $!\n";
open ( CLASSESFILE, ">$natural_class_filename") or die "Can't write classes file: $!\n";
#the first line contains the feature names
$line = <FEATUREFILE>;
$line = cleanup_line($line);
# also for some reason these often begin without tabs, but just in case
# they do have tabs
$line =~ s/^\s+//;

# now we just have feature names left
$number_of_features = (@features = split("\t",$line));
print "There are $number_of_features features in this file.\n";

my %feature_matrix;
my @segments;
my @list_of_segments;
my $number_of_segments = 0;

# Now we handle the translation of digraphs to single characters, if necessary.
# First, see if there's already a .trans file
if (-e $trans_filename) {
	$pre_existing_transfile = 1;
	print "Using pre-existing translation table: $trans_filename\n";
	open (TRANSFILE, "<$trans_filename");
	while ($line = <TRANSFILE>) {
		$transfile_line++;
		chomp($line);
		if ($line eq "") {
		    print "Warning: skipping blank line in $trans_filename (line $transfile_line)\n";
		} else {		
		    ($inputseg, $outputseg) = split(/\t/, $line);
		    if ($inputseg eq "" or $outputseg eq "") {
			    print "Warning! Ill-formed entry in line $transfile_line of translation table: [$inputseg] -> [$outputseg]\nPlease fix .trans file and try again.\n";
			    exit;
		    } elsif ($segs{$inputseg} ne undef) {
		        print "Warning! [$inputseg] in line $transfile_line of translation table is already defined as $segs{$inputseg}\nI'm ignoring the second entry.\n";
		    } elsif ($used_segs{$outputseg} ne undef) {
		        print "Warning! Line $transfile_line of translation file: the output symbol [$outputseg] is already being used for [$used_segs{$outputseg}]\nI'm going to assign a new arbitrary symbol to [$inputseg].\n(Feel free to edit the .trans file and rerun me.\n";
		    } else {
			    $list_of_segs[ $number_of_segments ] = $inputseg;
			    $segs{$inputseg} = $outputseg;
			    $used_segs{$outputseg} = $inputseg;
			    $number_of_segments++;
		    }
		}
		
	}	
} else {
	# unnecessary, but just to be clear
	$pre_existing_transfile = 0;
}

# Now, check the features file, finding any digraphs or unknown symbols that might be in use:
my $number_of_digraphs = 0;
my $arbitrary_symbol = 159;
my $number_of_extras = 0;
while ($line = <FEATUREFILE>) {
    $line = cleanup_line($line); 
    ($seg, @values) = split("\t",$line);
    $seg =~ s/ //g;
    if ($seg eq "") {
	next;
    }
    
	if ($pre_existing_transfile) {
		# If all goes well, the pre-existing translation table should already have this segment; we just
		#   need to ensure that it does, and squawk if not, and save to add to file.
		if (!exists $segs{$seg}) {
			# Oops, this one wasn't listed in the translations table.  Rather than dying, we'll just add it, to be nice.			
			$list_of_segs[ $number_of_segments ] = $seg;
			$extra_segs[ $number_of_extras ] = $seg;
			$number_of_segments++;
			$number_of_extras++;
			if (length($seg) == 1) {
				# not a digraph, it can just stand for itself
				$segs{$seg} = $seg;
			} else {
				FINDSYMBOL: {
					while (1) {		
						$arbitrary_symbol++;
						$new_trans = chr($arbitrary_symbol);
						if (!exists $used_segs{$new_trans}) {
							# we're clear to use this one as the symbol
							$segs{$seg} = $new_trans;
							$used_segs{$new_trans} = $seg;
							last;
						}
					}
				}
			}
			print "Warning: [$seg] is used in the feature file, but does not appear in $trans_filename.\nI'm going to add it to $trans_filename, with an arbitrary value [$segs{$seg}].\n";
		}
	}
	else { # No pre-existing translation table; we're creating our own, if need be.
		$list_of_segs[ $number_of_segments ] = $seg;
		$extra_segs[ $number_of_extras ] = $seg;
		$number_of_segments++;
		$number_of_extras++;

		if (length($seg) > 1) {
			# If it's a digraph, we need to assign it a new arbitrary symbol
			if ($number_of_digraphs == 0) {
				print "\n***************************************************************************\nIt appears that you are using digraphs\nI'm going to create an internal translation table to unitary symbols, and store it in the file $trans_filename.\nFeel free to edit this file and rerun me, if you prefer different symbols\n***************************************************************************\n\n";
			}
			$number_of_digraphs++;
			FINDSYMBOL2: {
				while (1) {		
					$arbitrary_symbol++;
					$new_trans = chr($arbitrary_symbol);
					if (!exists $used_segs{$new_trans}) {
						# we're clear to use this one as the symbol
						$segs{$seg} = $new_trans;
						$used_segs{$new_trans} = $seg;
						last;
					}
				}
			}
		} else {
			# a single segment; life is easy.
			$segs{$seg} = $seg;
			$used_segs{$seg} = $seg;
			# $list_of_segs .= "$seg ";
		}
	}
}    

if ( $number_of_extras > 0) {
	# We need to update the translation file to include whatever segments were not in it.
	#    (This may involve create a translation file, if there was none to begin with)
	print "Some segments appeared in the feature file that were not given pre-determined symbols in the .trans file.\nI'm going to assign them arbitrary symbols in the file $trans_filename  (This is harmless)\nFeel free to edit $trans_filename and rerun me.\n\n";
	open (TRANSFILE, ">$trans_filename") or die "Warning! Can't write to translation table file: $!\n";
	foreach my $i (@list_of_segs) {
#	    print "Outputting segment $i to .trans file \n";
		printf TRANSFILE "$i\t$segs{$i}\n";
	}
	close(TRANSFILE);
}

close(FEATUREFILE);
open( FEATUREFILE, "<$featurefile")  or die "Can't open input file: $!\n";
# We need to skip the first line, which contains just the names of the features:
$line = <FEATUREFILE>;
my $number_of_segments = 0;
READFILE: {
	print "Reading in feature values\n";
	
    while ($line = <FEATUREFILE>)
    {
	# clean up the line (remove quotes, extra white space, etc.)
	$line = cleanup_line($line); 

	# a line contains the segment name, possibly a code (which we'll ignore), and some feature values
	($seg, @values) = split("\t",$line);
	
	$seg =~ s/ //g;
	# Now we'll translate segments to their one-character representations
	$seg = $segs{$seg};

	# if the first value is nothing but digits or is completely empty, 
	# it's a code (assuming no scalar features!!!)
	if ($values[0] =~ /^\d*$/ or $values[0] eq " ") {
	   # then just get rid of it
	   shift(@values);	  
	}

	# ignore blank lines
	if (@values > 0) {
	    # store the segment in the list of segments, for convenience
	    $segments[$number_of_segments] = $seg;
	    for ($i = 0; $i < @values; $i++) {
		    $feature_matrix{$seg}[$i] = @values[$i];
	    }	
	    $number_of_segments++;   

	    # a consistency check on the rest of the line
	    if (@values ne $number_of_features) {
		print "Warning! segment $seg has an incorrect number of features \(".@values." instead of ".$number_of_features."\).\n";
	    }
	}
    }
}
close(FEATUREFILE);
# Finally, we can check that this has been read in right
print_matrix();

#Now its time to construct the matrix of strings for unitary classes.
#The structure for this part will be: @unitary_classes[$number_of_features][2],
# where the final array is [0] for negative (-) values, [1] for positive (+)
my @unitary_classes;
foreach $segment (sort keys %feature_matrix) {   
    for ($f = 0; $f < $number_of_features; $f++) {
	if ($feature_matrix{$segment}[$f] eq '+') {
	    # this segment has a positive value for this feature --
	    # include it in the unitary class [1] array
	    $unitary_classes[$f][1] .= $segment;

	} elsif ($feature_matrix{$segment}[$f] eq '-') {
	    # this segment has a negative value for the feature --
	    # include it in the unitary class [0] array
	    $unitary_classes[$f][0] .= $segment;
	} elsif ($feature_matrix{$segment}[$f] ne '0') {
	    print "Warning: ignoring illegal feature value, number $f for segment $segment: $feature_matrix{$segment}[$f]\n";
	}	
	# if 0, ignore
    }
}

# now @unitary_classes should have lists of unitary natural classes
# print it out to check
printf OUTPUTFILE "Unitary natural classes:\n";
for ($f=0; $f < $number_of_features; $f++) {
    for ($val = 1; $val >= 0; $val--) {
	printf OUTPUTFILE "\[";
	if ($val == 1) {
	    printf OUTPUTFILE "+";
	} else {
	    printf OUTPUTFILE "-";
	}
	printf OUTPUTFILE "$features[$f]\]: $unitary_classes[$f][$val]\n";
    }
}


# now that we have the unitary classes for all features, we can check which
# ones are unique, add them to the list of natural classes, and iterate from there
my $number_of_classes = 0;
for ($c = 0; $c < @unitary_classes; $c++) {
    # gotta consider both positive and negative polarity
    for ($polarity = 0; $polarity <= 1; $polarity++) {
	# eliminate spaces that creep in somehow
	$unitary_classes[$c][$polarity] =~ s/ //g;	
	
	# if the class already exists, it will have a length specified
	if ($descs{$unitary_classes[$c][ $polarity ]}[0] > 0) {
	    # we can ignore this one, we already saw it
#	    print "Skipping class \[";	
#	    if ($polarity == 1) {
#		print "+";
#	    } else {
#		print "-";
#	    }
#	    print "$features[$c]\] \($unitary_classes[$c][$polarity]\) because it is already described by \[$descs{$unitary_classes[$c][$polarity]}[1]\].\n";
	} else {
            # This clause eliminates the empty class: 
#	    if (length($unitary_classes[$c][$polarity]) eq 0) {
#		next;
#	    } else {
		# this one is new; install it as a class
		$classes[$number_of_classes] = $unitary_classes[$c][$polarity];	   
		$all_classes .= "\[$classes[$number_of_classes]\] ";	   
		
		# and add its description to the hash table %descs;
		# first the length
		$descs{$unitary_classes[$c][$polarity]}[0] = 1;	   
		# and now the feature combination
		if ($polarity == 1) {
		    $descs{$unitary_classes[$c][$polarity]}[1] .= "+";			
		} else {
		    $descs{$unitary_classes[$c][$polarity]}[1] .= "-";			
		}
		$descs{$unitary_classes[$c][$polarity]}[1] .= $features[$c];
		$number_of_classes++;
#	    } # end of clause eliminating the empty class	   	  
	}
    } # end of looping through both pos and neg
} # end of cycling through unitary classes

# printf OUTPUTFILE "\nAll classes:\n$all_classes\n\n";


$number_of_unitary_classes = $number_of_classes;

# OK, so far we have a set of natural classes that includes all of the unique unitary classes.
# The next step is to combine them to create classes with more complex descriptions.
# We'll go through the ones we already know and compare.

#print "Number of unitary classes = $number_of_unitary_classes\n";

# for keeping track of progress:
$number_considered = 0;
$percent_generalized = 0;
print "Discovering natural classes.  Percent done:       ";

# to get the updated percent to print often enough to be useful, 
# we have to make STDOUT "hot"
my $ofh = select STDOUT;
$| = 1;
select $ofh;

# the size of the matrix increases as we find more natural classes; this
# $exclude variable is solely for the purpose of getting an accurate estimate
# of how far we are through the task
$exclude = 0;
for ($c1 = 0; $c1 < $number_of_unitary_classes; $c1++) {
    # for each class, compare against all of the classes after it
#    print "Generalizing unitary class $c1; number of classes so far = $number_of_classes\n";   
    $old_number_of_classes = $number_of_classes;
    for ($c2 = $c1+1; $c2 < $number_of_classes; $c2++) {

	# Let's keep track of progress so we can report it
	$number_considered++;
	# This looks convoluted, but it's purely for morale: sometimes
	# a bunch of new classes are discovered and we find out that we have
	# more left to do than we thought.  Rather than making the percent
	# done report go *down* (like watching the clock tick backwards), 
	# we'll just lie and keep it constant for a while until we catch up.
	
	# the number of boxes we have to consider depends on the number
	# of classes that we have; this keeps updating.
	# since there are more classes than unitary classes, we have a
	# rectangular matrix to explore, and the math is sort of ugly

	$number_to_consider = plustorial($number_of_classes) - plustorial($number_of_classes-$number_of_unitary_classes) - $exclude;
	$real_percent_generalized = ($number_considered / $number_to_consider) * 100;
	if ($real_percent_generalized > $percent_generalized) {
	    $percent_generalized = $real_percent_generalized;
	    print chr(8)x6;
	    print sprintf("%4.f",$percent_generalized) . " \%";
	}

	$candidate_class = find_dups($classes[$c1].$classes[$c2]);
	# somehow a space seems to get inserted sometimes; kill it
	
	$candidate_class =~ s/ //g;

	# proceed only if intersection is > 0
	if (length($candidate_class) > 0) {	 
	    $candidate_desc = $descs{$classes[$c1]}[1].", ".$descs{$classes[$c2]}[1];	        	
	    
	    # Actually, when we are using overspecification, then sometimes the two classes being merged will redundantly
	    # have a feature specification in common; need to check for duplicates and weed them out
	    # Start by making an array
	    @candidate_desc = sort(split(", ",$candidate_desc));	    
	    if ($specification eq "over") {
		for (my $i = scalar (@candidate_desc) -1; $i >= 0; $i--) {
		    if ($candidate_desc[$i] eq $candidate_desc[$i-1]) {
#			print "Removing redundant $candidate_desc[$i] from $candidate_desc\n";
			splice(@candidate_desc, $i, 1);
#			print "\tResult = ".join(", ",@candidate_desc)." (length = $candidate_length)\n";
		    }
		}	    
		$candidate_desc = join(", ",@candidate_desc);	    
	    }
	   $candidate_length = scalar(@candidate_desc);
	    
	    # if we have already seen this class
	    if( 0 < $descs{$candidate_class}[0] ) {
	       # we've seen this class before; 
		# also, we might want to replace with new description, though, depending on the
	       # desired level of specification
		if ($specification eq "under") {
		    if ( $descs{$candidate_class}[0] > $candidate_length ) {
		     # we saw this one before, but this new description is shorter
#			print "Found a shorter way to describe the class \($candidate_class\): \[$candidate_desc\] instead of \[$descs{$candidate_class}[1]\]\n";			
			$descs{$candidate_class}[0] = $candidate_length;	       
			$descs{$candidate_class}[1] = $candidate_desc;
		    }
		} else { # overspecification is desired; take the longest description possible
		    if ( $descs{$candidate_class}[0] < $candidate_length ) {
			 # we saw this one before, but this new description is shorter
#			print "Found a longer way to describe the class \($candidate_class\): \[$candidate_desc\] instead of \[$descs{$candidate_class}[1]\]\n";			
			$descs{$candidate_class}[0] = $candidate_length;	       
			$descs{$candidate_class}[1] = $candidate_desc;
		    }		    
		}		
	    } elsif (0 < $candidate_length) {
		# otherwise, we haven't seen it
		$classes[$number_of_classes] = $candidate_class;	   
		$all_classes .= "\[$classes[$number_of_classes]\] ";	   
		$descs{$candidate_class}[0] = $candidate_length;	   
#		$descs{$candidate_class}[1] = $descs{$classes[$c1]}[1].", ".$descs{$classes[$c2]}[1];	   
		$descs{$candidate_class}[1] = $candidate_desc;	   
		$number_of_classes++;	    	   
	    }
	} # end if candidate_class is non-empty
    }
    # Now done with an inner loop; we've discovered some new natural
    # classes which we can exclude from our future estimates of the
    # amount of work that has to be done (this is boring and complicated,
    # but is necessary for the "Percent done" estimate to turn out correct)
    $exclude += (($number_of_classes - $old_number_of_classes) * $c1);
}

# purely aesthetic: change final percent done to 100%
print chr(8)x6;
print sprintf("%4.f",100) . " \%";

print "\n\n";

# finally, add the "null description" class, which contains all of the segments
foreach $segment (sort keys %feature_matrix) {
    $classes[$number_of_classes] .= $segment;       
}
$descs{$classes[$number_of_classes]}[0] = 0;
$descs{$classes[$number_of_classes]}[1] = "+seg";
# the description of the last class has a length of 0, and no feature specification
$descs{$classes[$number_of_classes]}[0] = 0;
$descs{$classes[$number_of_classes]}[1] = "+seg";
$all_classes .= "\[$classes[$number_of_classes]\] ";	   
$number_of_classes++;

# Finally, need to protect $all_classes in case it has special characters in it
$all_classes = "\Q".$all_classes."\E";

# Now that we have all the classes compiled, it's intuitive to look at them in order of the number of segments included
@classes = sort {length($a) <=> length($b)} @classes;

# Generally, one would like their feature system to provide a unique description for each segment in the
#   inventory; but with privative features, this can sometimes be tricky (it's easy to accidentally distinguish a
#   phoneme with a privative feature, and leave it's "unmarked" counterpart undescribable.
# So, we perform a check to make sure that all segments are describable.
print OUTPUTFILE "---------------------------------------------------------\nOptimal descriptions of each phoneme using features:\n";
for ($i = 0; $i < $number_of_segments; $i++) {
    $segment = $segments[$i];
    print OUTPUTFILE "[$used_segs{$segment}]\t";
#    if ($all_classes =~ /\[$segment\]/ {
    if (defined $descs{$segment}) {
	print OUTPUTFILE "$descs{$segment}[1]\n";
    } else {
		print OUTPUTFILE "***\n";
		print "Warning: segment [$used_segs{$segment}] is not uniquely describable using this feature set.\n";
		# Find the shortest class that contains the segment; since  @classes is sorted by length, one way is to
		#   find the first occurence in @classes
		for (my $c = 0; $c < $number_of_classes; $c++) {
			$class = $classes[$c];
			if ($class =~ /\Q$segment\E/) {	    
				$description = $descs{$class}[1];
				$class =~ s/\Q$segment\E//;
				# $class =~ s/\\//g;
				@class_segs = split( "", $class);
				$print_class = "";
				foreach $class_seg (@class_segs) {
					$print_class .= "$used_segs{$class_seg},";
				}
				$print_class =~ s/,$//g;
				
				print "\tSegment does not contrast with [$print_class]\n\tFeatures: [$description]\n\n";
				last;
			}
		}
    }
}

# now that we've done that, let's print it out and see what we've got
printf OUTPUTFILE "---------------------------------------------------------\nAll natural classes:\n";
for ($c = 0; $c < $number_of_classes; $c++) {
    printf OUTPUTFILE "$classes[$c]\t$descs{$classes[$c]}[0]\t$descs{$classes[$c]}[1]\n";
    printf CLASSESFILE "$classes[$c]\t$descs{$classes[$c]}[0]\t$descs{$classes[$c]}[1]\n";
}
printf OUTPUTFILE "---------------------------------------------------------\n";
print "Number of natural classes: $number_of_classes\n";
print "\nReady to calculate similarities.  Should I calculate just the pairwise similarity of individual segments, or the similarity of segments to all natural classes?\n\t1. Just pairwise segmental similarity\n\t2. Similarity to all classes (Warning: may be very slow!)\n? ";

while (1) {
    $desired = <STDIN>;
    chomp($desired);
    if ($desired eq "1" or $desired eq "1.") {
	$desired = "segs";
	last;
    } elsif ($desired eq "2" or $desired eq "2.") {
	$desired = "all";
	last;
    } else {
	print "Sorry, couldn't understand response.  Please enter 1 or 2.\n? ";
    }
}

# Start with the pairwise comparisons
# First, make a header row for the similarity matrix file
for (my $seg = 0; $seg < $number_of_segments; $seg++) {
	# We'll translate back into the same values as the original features file
    printf SIMMATRIXFILE "\t$used_segs{$segments[$seg]}";   
}
printf SIMMATRIXFILE "\n";

# And the same thing in list form in the SimTab file
open ( SIMTABFILE, ">$similarity_table_filename") or die "Can't write similarity table file: $!\n";
# Start with a header row for the similarity table file
	if ($desired eq "all") {
    		printf SIMTABFILE "class1\tclass2\tshared\ttotal\tsimilarity\n";
	} else {
    		printf SIMTABFILE "seg1\tseg2\tshared\ttotal\tsimilarity\n";
	}

for (my $seg1 = 0; $seg1 < $number_of_segments; $seg1++) {
    printf SIMMATRIXFILE "$used_segs{$segments[$seg1]}";   

    for (my $seg2 = 0; $seg2 <= $seg1; $seg2++) {
	if ($seg1 eq $seg2) {
	    printf SIMMATRIXFILE "\t1";
	} else {
	    @compare = sort($segments[$seg1], $segments[$seg2]);
	    # compose the target string
	    $target_string = "\\\[\[^\\\]\]*";	   
	    for ($i = 0; $i < @compare; $i++) {
		my $segment = $compare[$i];
		# for some reason, extra spaces creep in sometimes when
		# the segment is a metacharacter (like question mark)
		$segment =~ s/ //g;
		# and we should protect the metacharacters
		$segment =~ s/(\W)/\\$1/g;
		$target_string .= "$segment\[^\\\]]*";	
	    }	   
	    $target_string .= "\\\]";

	    # count shared classes
	    $shared_classes = 0;	    	   
	    while ($all_classes =~ m/($target_string)/g) {
		# yet another relevant class
		$shared_classes++;			
	    }

#	    if ($shared_classes <= 1) {
#		print "The segments \[$segments[$seg1]\] and \[$segments[$seg2]\] (target string: $target_string) occur in only 1 natural classes.\n";
#	    }

	    # now count non-shared classes.  non-shared = (classes seg1 + classes seg2) - shared
	    $total_classes = 0;	   
	    # We need to count the classes that contain seg1 and the classes that contain seg2.
	    # First, seg1
	    $segment1 = $segments[$seg1];
	    $segment1 =~ s/ //g;
	    $segment1 =~ s/(\W)/\\$1/g;
	    $target_string = "\\\[\[^\\\]\]*" . $segment1 . "\[^\\\]]*\\\]";
	    while ($all_classes =~ m/($target_string)/g) {
		$total_classes++;			
	    }
	    # And then seg2
	    $segment2 = $segments[$seg2];
	    $segment2 =~ s/ //g;
	    $segment2 =~ s/(\W)/\\$1/g;
	    $target_string = "\\\[\[^\\\]\]*" . $segment2 . "\[^\\\]]*\\\]";
	    while ($all_classes =~ m/($target_string)/g) {
		$total_classes++;			
	    }
	    $total_classes = $total_classes - $shared_classes;	    	   

	    # One other small issue: whether to count the "+seg" class
	    # each time (that is, the superset class that contains all segments)
	    # Frisch, Broe and Pierrehumbert imply that we should, but the
	    # Shared/Unshared counts from Kie's similarity program imply that
	    # she does not.  Doesn't matter much because it will just add or
	    # substract 1 from both the numerator and denominator of all comparisons.


	    # For consistency with Kie's program, subtract if desired.
	    if ($superclass eq "exclude") {
		$total_classes -= 1;
		$shared_classes -= 1;
	    }

	    $similarity = $shared_classes / $total_classes;	   
	    printf SIMMATRIXFILE "\t%.4f", $similarity;
	    printf SIMTABFILE "$segments[$seg1]\t$segments[$seg2]\t$shared_classes\t$total_classes\t%.3f\n", $similarity;	    
	}
    }
    printf SIMMATRIXFILE "\n";
}
close (SIMMATRIXFILE);

if ($desired eq "all") {

print "Calculating similarities of segments to all classes.  Percent done:       ";
$number_considered = 0;
#$number_to_consider = plustorial($number_of_classes);
$number_to_consider = $number_of_classes * $number_of_segments;

    
# 7/22/03  Previously, this was done by cycling through all of the known classes,
#    and comparing each class to all other classes with length 1.
#    The problem with this approach was that if a particular segment in the
#    inventory does not constitute its own natural class (because of privativity,
#    for example), then its individual similarity to other classes got left out
#    of the similarity table.
#    What we really need to do here is to compare two things:
#    (1) The similarity of each segment to each segment (already done above for the similarity matrix)
#    (2) The similarity of each class to each segment.
        
# we'll make matrices of shared and not-shared classes
    for ($c1 = 0; $c1 < $number_of_classes; $c1++) {
	# We don't need similarity to the empty class, or to single-segment classes
	#  (since the latter were already taken care of above)
	next if (length($classes[$c1]) eq 0 or length($classes[$c1]) eq 1);
	
	# For each class, compare all the segments
      CALCSIM:
	  for (my $seg = 0; $seg < $number_of_segments; $seg++) {
	    $number_considered++;
	    $percent_done = ($number_considered / $number_to_consider) * 100;
	    print chr(8)x6;
	    print sprintf("%4.f",$percent_done) . " \%";
	    
	    # if we don't compare generalized with generalized, we don't need to 
	    # compare multiple classes;  just segs against other segs and 
	    # segs against classes

	    # get the segments in alphabetical order and remove duplicates
	    my $intersection = remove_dups($classes[$c1].$segments[$seg]);
	    @compare = split('',$intersection);	   
	    
	    # now compose the target string
	    $target_string = "\\\[\[^\\\]\]*";	   
	    for ($i = 0; $i < @compare; $i++) {
		$target_string .= "\Q$compare[$i]\E\[^\\\]]*";	
	    }	   
	    $target_string .= "\\\]";
	    # count shared classes
	    $shared_classes = 0;	    	   
	    while ($all_classes =~ m/($target_string)/g) {
		# yet another relevant class
		$shared_classes++;			
	    }
	    
	    # now count non-shared classes.  non-shared = (classes c1 + classes c2) - shared
	    $total_classes = 0;	   
	    # We need to count the classes that contain c1 and the classes that contain c2.
	    # Unfortunately, this is more cumbersome than just counting classes that contain single segments,
	    # because classes that contain another natural class are not those that contain *any* of its elements,
	    # but rather those that contain *all* of its elements  (e.g., [eio] doesn't contain [iu], but [eiou] does)
	    
	    @c1 = split('',$classes[$c1]);	   
	    $target_string = "\\\[\[^\\\]\]*";	   
	    for ($i = 0; $i < @c1; $i++) {
		$target_string .= "\Q$c1[$i]\E\[^\\\]]*";	
	    }	   
	    $target_string .= "\\\]";
	    while ($all_classes =~ m/($target_string)/g) {
		$total_classes++;			
	    }
	    @c2 = split('',$segments[$seg]);	   
	    $target_string = "\\\[\[^\\\]\]*";	   
	    for ($i = 0; $i < @c2; $i++) {
		$target_string .= "\Q$c2[$i]\E\[^\\\]]*";	
	    }	   
	    $target_string .= "\\\]";
	    while ($all_classes =~ m/($target_string)/g) {
		$total_classes++;			
	    }

	    $total_classes = $total_classes - $shared_classes;	    	   

	    # Again, for the sake of consistency with Kie's program
	    if ($superclass eq "exclude") {
		$total_classes -= 1;
		$shared_classes -=1;
	    }


	    $shared_classes[$c1][$seg] = $shared_classes;	   
	    $total_classes[$c1][$seg] = $total_classes;  
	    $similarity[$c1][$seg] = $shared_classes / $total_classes;
	    printf SIMTABFILE "$classes[$c1]\t$segments[$seg]\t$shared_classes[$c1][$seg]\t$total_classes[$c1][$seg]\t%.3f\n", $similarity[$c1][$seg];
	}

    }
# purely aesthetic: change final percent done to 100%
print chr(8)x6;
print sprintf("%4.f",100) . " \%";

print "\nDone calculating similarities.\n\n";
}


close(OUTPUTFILE);
close(SIMTABFILE);
#############################################################################
sub cleanup_line {
     $clean = shift(@_);         
     chomp($clean);
     #Kie's version of these files has fields in quotes and with commas, "X,"
     $clean =~ s/\"([^\"]*),\"/$1/g;
     $clean =~ s/,(\s|$)/$1/g;
     # get rid of syntactic elements like "No more features" or "EndOfLine"
     $clean =~ s/(\"?NoMoreFeatures\"?|NoMoreSegments|EndOfLine)//g;
     # also, sometimes there are extra white spaces at the ends of lines
     $clean =~ s/\s+$//;
     return $clean;
}
#############################################################################
sub find_dups {
    # this subroute takes a string and returns just the characters that occur twice
    my @input = sort split('',shift(@_));
    my $output;   
    for ($i = 1; $i < @input; $i++) {
	if ($input[$i] eq $input[$i-1]) {
	    # a match; add to the output if it isn't already there
	    if ($output =~ /\Q$input[$i]\E/) {
		# do nothing
	    } else {
	        # add it
	        $output .= $input[$i];	        	
	    }
	}
    }
    return $output;       
}
#############################################################################
sub remove_dups {
    my $clean = join('',sort split('',shift(@_)));
    $clean =~ s/(.)\1+/$1/g;
    return $clean;   
}

#############################################################################
sub print_matrix {
   printf OUTPUTFILE "---------------------FEATURE MATRIX----------------------\n";
   # first print the feature names
    printf OUTPUTFILE "\t".join("\t",@features)."\n";   
    for ($i = 0; $i < $number_of_segments; $i++) {
	printf OUTPUTFILE "$segments[$i]";
	for ($f = 0; $f < $number_of_features; $f++) {
	    printf OUTPUTFILE "\t$feature_matrix{$segments[$i]}[$f]";
	}
	printf OUTPUTFILE "\n";
    }
   printf OUTPUTFILE "\n---------------------------------------------------------\n";
}

#############################################################################
sub plustorial {
  my ($val) = @_;
  return(1) if $val<2;
  return($val + plustorial($val-1));
}
